/**
 * FreeRDP: A Remote Desktop Protocol Implementation
 *
 * Copyright 2014 Marc-Andre Moreau <marcandre.moreau@gmail.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#ifndef FREERDP_SERVER_SHADOW_WIN_DXGI_H
#define FREERDP_SERVER_SHADOW_WIN_DXGI_H

#if _WIN32_WINNT >= 0x0602
//#define WITH_DXGI_1_2	1
#endif

#ifdef WITH_DXGI_1_2

#ifndef CINTERFACE
#define CINTERFACE
#endif

#include <D3D11.h>
#include <dxgi1_2.h>

#endif

#include "win_shadow.h"

#ifdef __cplusplus
extern "C" {
#endif

#ifdef WITH_DXGI_1_2

int win_shadow_dxgi_init(winShadowSubsystem* subsystem);
int win_shadow_dxgi_uninit(winShadowSubsystem* subsystem);

int win_shadow_dxgi_fetch_frame_data(winShadowSubsystem* subsystem,
	BYTE** ppDstData, int* pnDstStep, int x, int y, int width, int height);

int win_shadow_dxgi_get_next_frame(winShadowSubsystem* subsystem);
int win_shadow_dxgi_get_invalid_region(winShadowSubsystem* subsystem);

#endif

#ifdef __cplusplus
}
#endif

#endif /* FREERDP_SERVER_SHADOW_WIN_DXGI_H */
